import numpy as np
import torch


class CnnNet(torch.nn.Module):
    def __init__(self, in_frames=3):
        super(CnnNet, self).__init__()

        self.conv1 = torch.nn.Conv2d(in_frames, 16, kernel_size=4, stride=2, padding=1)
        self.bn1 = torch.nn.BatchNorm2d(16)
        self.conv2 = torch.nn.Conv2d(16, 32, kernel_size=4, stride=2, padding=1)
        self.bn2 = torch.nn.BatchNorm2d(32)
        self.conv3 = torch.nn.Conv2d(32, 64, kernel_size=3, stride=2, padding=0)
        self.bn3 = torch.nn.BatchNorm2d(64)
        self.conv4 = torch.nn.Conv2d(64, 32, kernel_size=3, stride=1, padding=0)
        self.bn4 = torch.nn.BatchNorm2d(32)
        self.phi = torch.relu
        self.drop = torch.nn.Dropout(p=0.4)

    def forward(self, x):
        h = self.phi(self.bn1(self.conv1(x)))
        h = self.phi(self.bn2(self.conv2(h)))
        h = self.phi(self.bn3(self.conv3(h)))
        h = self.drop(self.phi(self.bn4(self.conv4(h))))
        return h


class CnnLstmNet(torch.nn.Module):
    def __init__(self, cnn, lstm_hidden_dim, lstm_layer_dim):
        super(CnnLstmNet, self).__init__()

        self.in_dim = 288
        self.cnn = cnn
        self.hidden_dim = lstm_hidden_dim
        self.layer_dim = lstm_layer_dim
        self.lstm = torch.nn.LSTM(self.in_dim, self.hidden_dim, self.layer_dim,
                                  batch_first=True, dropout=0.4)

    def forward(self, x):
        b, s, c, h, w = x.shape

        # Initialize hidden state with zeros
        h0 = torch.zeros(self.layer_dim, x.shape[0], self.hidden_dim,
                         dtype=x.dtype, layout=x.layout, device=x.device).requires_grad_()
        # Initialize cell state
        c0 = torch.zeros(self.layer_dim, x.shape[0], self.hidden_dim,
                         dtype=x.dtype, layout=x.layout, device=x.device).requires_grad_()

        # reshape time to batch dimension
        x = torch.reshape(x, shape=(b * s, c, h, w))
        # apply cnn encoder
        x = self.cnn(x)
        # split time and batch dimension
        x = torch.reshape(x, shape=(b, s, -1))

        return self.lstm(x, (h0.detach(), c0.detach()))


class ActionNet(torch.nn.Module):
    def __init__(self, lstm_hidden_dim, action_out_dim):
        super(ActionNet, self).__init__()

        self.out_dim = action_out_dim
        self.hidden_dim = lstm_hidden_dim
        self.final = torch.nn.Linear(self.hidden_dim, self.out_dim)

    def forward(self, x):
        return self.final(x)


class CameraNet(torch.nn.Module):
    def __init__(self, lstm_hidden_dim, camera_out_dim):
        super(CameraNet, self).__init__()

        self.out_dim = camera_out_dim
        self.hidden_dim = lstm_hidden_dim
        self.final = torch.nn.Linear(self.hidden_dim, self.out_dim)

    def forward(self, x):
        return self.final(x)


class ValueNet(torch.nn.Module):
    def __init__(self, lstm_hidden_dim):
        super(ValueNet, self).__init__()

        self.hidden_dim = lstm_hidden_dim
        self.final = torch.nn.Linear(self.hidden_dim, 1)

    def forward(self, x):
        return self.final(x)


class EquipNet(torch.nn.Module):
    def __init__(self, lstm_hidden_dim, out_dim):
        super(EquipNet, self).__init__()

        self.out_dim = out_dim
        self.hidden_dim = lstm_hidden_dim
        self.final = torch.nn.Linear(self.hidden_dim, self.out_dim)

    def forward(self, x):
        return self.final(x)


class PlaceNet(torch.nn.Module):
    def __init__(self, lstm_hidden_dim, out_dim):
        super(PlaceNet, self).__init__()

        self.out_dim = out_dim
        self.hidden_dim = lstm_hidden_dim
        self.final = torch.nn.Linear(self.hidden_dim, self.out_dim)

    def forward(self, x):
        return self.final(x)


class CraftNet(torch.nn.Module):
    def __init__(self, lstm_hidden_dim, out_dim):
        super(CraftNet, self).__init__()

        self.out_dim = out_dim
        self.hidden_dim = lstm_hidden_dim
        self.final = torch.nn.Linear(self.hidden_dim, self.out_dim)

    def forward(self, x):
        return self.final(x)


class NearbyCraftNet(torch.nn.Module):
    def __init__(self, lstm_hidden_dim, out_dim):
        super(NearbyCraftNet, self).__init__()

        self.out_dim = out_dim
        self.hidden_dim = lstm_hidden_dim
        self.final = torch.nn.Linear(self.hidden_dim, self.out_dim)

    def forward(self, x):
        return self.final(x)


class NearbySmeltNet(torch.nn.Module):
    def __init__(self, lstm_hidden_dim, out_dim):
        super(NearbySmeltNet, self).__init__()

        self.out_dim = out_dim
        self.hidden_dim = lstm_hidden_dim
        self.final = torch.nn.Linear(self.hidden_dim, self.out_dim)

    def forward(self, x):
        return self.final(x)


class Net(torch.nn.Module):
    def __init__(self, in_frames=3, lstm_hidden_dim=256, lstm_layer_dim=1, action_out_dim=8, camera_out_dim=2,
                 equip_dim=8, place_dim=7, craft_dim=5, nearbycraft_dim=8, nearbysmelt_dim=3):
        super(Net, self).__init__()
        self.cnn = CnnNet(in_frames=in_frames)
        self.cnn_lstm = CnnLstmNet(self.cnn, lstm_hidden_dim=lstm_hidden_dim, lstm_layer_dim=lstm_layer_dim)
        self.action_net = ActionNet(lstm_hidden_dim=lstm_hidden_dim, action_out_dim=action_out_dim)
        self.camera_net = CameraNet(lstm_hidden_dim=lstm_hidden_dim, camera_out_dim=camera_out_dim)

        self.equip_net = EquipNet(lstm_hidden_dim=lstm_hidden_dim, out_dim=equip_dim)
        self.place_net = PlaceNet(lstm_hidden_dim=lstm_hidden_dim, out_dim=place_dim)
        self.craft_net = CraftNet(lstm_hidden_dim=lstm_hidden_dim, out_dim=craft_dim)
        self.nearbycraft_net = NearbyCraftNet(lstm_hidden_dim=lstm_hidden_dim, out_dim=nearbycraft_dim)
        self.nearbysmelt_net = NearbySmeltNet(lstm_hidden_dim=lstm_hidden_dim, out_dim=nearbysmelt_dim)

        self.cnn_val = CnnNet(in_frames=in_frames)
        self.cnn_lstm_val = CnnLstmNet(self.cnn_val, lstm_hidden_dim=lstm_hidden_dim, lstm_layer_dim=lstm_layer_dim)
        self.value_net = ValueNet(lstm_hidden_dim=lstm_hidden_dim)

    def policy_parameters(self):
        return list(self.cnn_lstm.parameters()) + \
               list(self.action_net.parameters()) + list(self.camera_net.parameters())

    def value_parameters(self):
        return list(self.cnn_lstm_val.parameters()) + list(self.value_net.parameters())

    def _shared_parameters(self):
        return []

    def _set_bias_sigmoid(self, traj_stats, act_stats, net):
        traj_thresh = 0.3
        eps = 1e-12
        biases = np.full_like(act_stats, fill_value=-1000.0)
        biases[traj_stats >= traj_thresh] = -np.log(1 / act_stats[traj_stats >= traj_thresh] - 1 + eps)
        net.final.bias.data.copy_(torch.tensor(biases).to(net.final.bias.device))

    def _set_bias_softmax(self, traj_stats, act_stats, net):
        traj_thresh = 0.3
        eps = 1e-12
        biases = np.full_like(act_stats, fill_value=-1000.0)
        biases[traj_stats >= traj_thresh] = np.log(act_stats[traj_stats >= traj_thresh] + eps)
        net.final.bias.data.copy_(torch.tensor(biases).to(net.final.bias.device))

    def set_action_statistics(self, action_statistics):
        """Configure action heads according to action statistics from demonstrations.
        Set initial biases according to action frequencies and disable output heads
        with constant value in the demonstrations (i.e. disable heads for enum actions with
         constant 'none' prediction) for increased efficiency."""
        trajectory_statistics, frame_statistics = action_statistics
        t_binary_action_stats, t_craft_stats, t_equip_stats, \
        t_nearbycraft_stats, t_nearbysmelt_stats, t_place_stats = trajectory_statistics
        a_binary_action_stats, a_craft_stats, a_equip_stats, \
        a_nearbycraft_stats, a_nearbysmelt_stats, a_place_stats = frame_statistics

        self._set_bias_sigmoid(t_binary_action_stats, a_binary_action_stats, self.action_net)
        self._set_bias_softmax(t_craft_stats, a_craft_stats, self.craft_net)
        self._set_bias_softmax(t_equip_stats, a_equip_stats, self.equip_net)
        self._set_bias_softmax(t_nearbycraft_stats, a_nearbycraft_stats, self.nearbycraft_net)
        self._set_bias_softmax(t_nearbysmelt_stats, a_nearbysmelt_stats, self.nearbysmelt_net)
        self._set_bias_softmax(t_place_stats, a_place_stats, self.place_net)

    def forward(self, x):
        h = self.cnn_lstm(x)[0][:, -1, ...]
        action_out = self.action_net(h)
        camera_out = self.camera_net(h)

        equip_out = self.equip_net(h)
        place_out = self.place_net(h)
        craft_out = self.craft_net(h)
        nearbycraft_out = self.nearbycraft_net(h)
        nearbysmelt_out = self.nearbysmelt_net(h)

        h = self.cnn_lstm_val(x)[0][:, -1, ...]
        value_out = self.value_net(h)

        return action_out, camera_out, equip_out, place_out, craft_out, nearbycraft_out, nearbysmelt_out, value_out


class BaseNetwork(Net):

    def __init__(self, ):
        super().__init__(lstm_hidden_dim=256, lstm_layer_dim=2, action_out_dim=8, camera_out_dim=22,
                         equip_dim=8, place_dim=7, craft_dim=5, nearbycraft_dim=8, nearbysmelt_dim=3)


class Network(BaseNetwork):
    def __init__(self, bias_crafting_actions=True):
        super().__init__()

        if bias_crafting_actions:
            binary_action_stats = np.zeros(8)
            binary_action_stats[0] = 1
            craft_stats = np.zeros(5)
            craft_stats[0] = 1
            equip_stats = np.zeros(8)
            equip_stats[0] = 1
            nearbycraft_stats = np.zeros(8)
            nearbycraft_stats[0] = 1
            nearbysmelt_stats = np.zeros(3)
            nearbysmelt_stats[0] = 1
            place_stats = np.zeros(7)
            place_stats[0] = 1
            action_statistics = binary_action_stats, craft_stats, equip_stats, nearbycraft_stats, nearbysmelt_stats, place_stats
            self.set_action_statistics([action_statistics, action_statistics])

    def forward(self, x):
        input_dict = x
        pov = input_dict['pov']

        logits, camera, equip, place, craft, nearbyCraft, nearbySmelt, value = super().forward(pov)

        return {"logits": logits, "camera": camera, "value": value,
                "act_equip": equip, "act_place": place, "act_craft": craft,
                "act_nearbyCraft": nearbyCraft, "act_nearbySmelt": nearbySmelt}


if __name__ == "__main__":
    """ main """
    in_frames = torch.from_numpy(np.random.randn(1, 10, 3, 48, 48).astype(np.float32))
    # binary_actions = torch.from_numpy(np.random.randn(1, 10, 9).astype(np.float32))
    # camera_actions = torch.from_numpy(np.random.randn(1, 10, 2).astype(np.float32))
    # equip_actions = torch.from_numpy(np.random.randn(1, 10, 8).astype(np.float32))
    # place_actions = torch.from_numpy(np.random.randn(1, 10, 7).astype(np.float32))
    # craft_actions = torch.from_numpy(np.random.randn(1, 10, 5).astype(np.float32))
    # nearbyCraft_actions = torch.from_numpy(np.random.randn(1, 10, 8).astype(np.float32))
    # nearbySmelt_actions = torch.from_numpy(np.random.randn(1, 10, 3).astype(np.float32))

    net = Network()
    out = net({"pov": in_frames
               # ,
               # "binary_actions": binary_actions, "camera_actions": camera_actions,
               # "enum_equip": equip_actions, "enum_place": place_actions, "enum_craft": craft_actions,
               # "enum_nearbyCraft": nearbyCraft_actions, "enum_nearbySmelt": nearbySmelt_actions
               })
    print("out_shape", out["logits"].shape, out["camera"].shape)
    print("equip", out["act_equip"].shape)
    actor_critic_params = list(net.parameters())
    print(len(actor_critic_params))
    policy_params = list(net.policy_parameters())
    print(len(policy_params))
    critic_parameters = set(actor_critic_params).difference(set(policy_params))
    print(len(critic_parameters))
